from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Sequence, Union

import lark
from lark import Lark, ParseTree
from lark.exceptions import VisitError

from heimdallm.bifrost import Bifrost as _BaseBifrost
from heimdallm.bifrosts.sql import exc
from heimdallm.llm import LLMIntegration
from heimdallm.llm_providers.mock import EchoMockLLM

if TYPE_CHECKING:
    import heimdallm.bifrosts.sql.envelope
    import heimdallm.bifrosts.sql.validator

from .envelope import TestSQLPromptEnvelope
from .visitors.ambiguity import AmbiguityResolver


class Bifrost(_BaseBifrost, ABC):
    """
    An abstract Bifrost for traversing SQL ``SELECT`` queries. This is used by the
    different SQL dialects.

    :param llm: The LLM integration to use.
    :param prompt_envelope: The prompt envelope used to wrap the untrusted human input
        and unwrap the untrusted LLM output.
    :param constraint_validators: A sequence of constraint validators that will be used
        to validate the parse tree returned by the ``tree_producer``. Only one validator
        needs to succeed for validation to pass.
    """

    # for tests
    @classmethod
    def mocked(
        cls,
        constraint_validators: Union[
            "heimdallm.bifrosts.sql.validator.ConstraintValidator",
            Sequence["heimdallm.bifrosts.sql.validator.ConstraintValidator"],
        ],
    ):
        """A convenience method for our tests. This creates a Bifrost that assumes its
        untrusted input is a SQL query already, so it does not need to communicate with
        the LLM, only parse and validate it.

        :param constraint_validators: A constraint validator or sequence of constraint
            validators to run on the untrusted input.

        :meta private:
        """
        if not isinstance(constraint_validators, Sequence):
            constraint_validators = [constraint_validators]

        llm = EchoMockLLM()
        return cls(
            llm=llm,
            constraint_validators=constraint_validators,
            prompt_envelope=TestSQLPromptEnvelope(
                llm=llm,
                db_schema="<schema>",  # doesn't matter
                validators=constraint_validators,
            ),
        )

    def __init__(
        self,
        *,
        llm: LLMIntegration,
        prompt_envelope: "heimdallm.bifrosts.sql.envelope.PromptEnvelope",
        constraint_validators: Sequence[
            "heimdallm.bifrosts.sql.validator.ConstraintValidator"
        ],
    ):
        super().__init__(
            llm=llm,
            prompt_envelope=prompt_envelope,
            grammar=self.build_grammar(),
            tree_producer=self.build_tree_producer(),
            constraint_validators=constraint_validators,
        )

    @classmethod
    @abstractmethod
    def reserved_keywords(cls) -> set[str]:
        """
        Returns the reserved keywords for the SQL dialect. Must be implemented in the
        subclass.

        :return: The reserved keywords.
        :meta private:
        """
        raise NotImplementedError

    @classmethod
    def build_tree_producer(cls) -> Callable[[Lark, str], ParseTree]:
        """
        Produces a that can create a single parse tree. May be implemented in a subclass
        if you want to do custom ambiguity resolution.

        :return: The parse tree producer.
        :meta private:
        """

        def parse(grammar: Lark, untrusted_query: str) -> ParseTree:
            ambig_tree = grammar.parse(untrusted_query)
            try:
                final_tree = AmbiguityResolver(
                    untrusted_query,
                    cls.reserved_keywords(),
                ).transform(ambig_tree)
            except VisitError as e:
                if isinstance(e.orig_exc, exc.BaseException):
                    raise e.orig_exc
                raise e
            return final_tree

        return parse

    @staticmethod
    @abstractmethod
    def build_grammar() -> Lark:
        """
        Produces the grammar that will be used to parse the dialect of SQL. Must be
        implemented in a subclass.

        :return: The Lark grammar for the SQL dialect.
        :meta private:
        """
        raise NotImplementedError

    def parse(self, untrusted_llm_output: str) -> ParseTree:
        """Parse the unwrapped SQL query from the LLM's output. Raise a SQL-specific
        exception if the query is not valid.

        :param untrusted_llm_output: The output from the LLM, which should be a SQL
            query. If it isn't, then our
            :meth:`heimdallm.bifrosts.sql.envelope.PromptEnvelope.unwrap` method failed.
        :raises InvalidQuery: If the query is not valid.
        :return: The Lark parse tree for the query.

        :meta private:
        """
        try:
            return super().parse(untrusted_llm_output)
        except lark.exceptions.UnexpectedEOF as e:
            raise exc.InvalidQuery(query=untrusted_llm_output) from e
        except lark.exceptions.UnexpectedCharacters as e:
            raise exc.InvalidQuery(query=untrusted_llm_output) from e
        except exc.BaseException as e:
            raise e
        except Exception as e:
            raise exc.InvalidQuery(query=untrusted_llm_output) from e



safe_date_functions = {
    "date",
    "time",
    "datetime",
    "julianday",
    "unixepoch",
    "strftime",
}

safe_string_functions = {
    "char",
    "format",
    "glob",
    "hex",
    "instr",
    "length",
    "like",
    "lower",
    "ltrim",
    "max",
    "min",
    "printf",
    "quote",
    "randomblob",
    "replace",
    "rtrim",
    "soundex",
    "substr",
    "substring",
    "trim",
    "unhex",
    "unicode",
    "upper",
    "zeroblob",
}

safe_math_functions = {
    "abs",
    "acos",
    "acosh",
    "asin",
    "asinh",
    "atan",
    "atan2",
    "atanh",
    "ceil",
    "ceiling",
    "cos",
    "cosh",
    "degrees",
    "exp",
    "floor",
    "ln",
    "log",
    "log10",
    "log2",
    "mod",
    "pi",
    "pow",
    "power",
    "radians",
    "random",
    "round",
    "sin",
    "sinh",
    "sign",
    "sqrt",
    "tan",
    "tanh",
    "trunc",
}


safe_misc_functions = {
    "coalesce",
    "ifnull",
    "iff",
    "nullif",
    "typeof",
}


safe_json_functions = {
    "json",
    "json_array",
    "json_array_length",
    "json_error_position",
    "json_extract",
    "json_insert",
    "json_object",
    "json_patch",
    "json_remove",
    "json_replace",
    "json_set",
    "json_type",
    "json_valid",
    "json_quote",
}

safe_agg_functions = {
    "avg",
    "count",
    "group_concat",
    "max",
    "min",
    "sum",
    "total",
}

safe_functions = (
    safe_date_functions
    | safe_string_functions
    | safe_math_functions
    | safe_misc_functions
    | safe_json_functions
    | safe_agg_functions
)

# we ensure that no unquoted identifiers match these keywords
reserved_keywords = {
    "abort",
    "action",
    "add",
    "after",
    "all",
    "alter",
    "always",
    "analyze",
    "and",
    "as",
    "asc",
    "attach",
    "autoincrement",
    "before",
    "begin",
    "between",
    "by",
    "cascade",
    "case",
    "cast",
    "check",
    "collate",
    "column",
    "commit",
    "conflict",
    "constraint",
    "create",
    "cross",
    "current",
    "current_date",
    "current_time",
    "current_timestamp",
    "database",
    "default",
    "deferrable",
    "deferred",
    "delete",
    "desc",
    "detach",
    "distinct",
    "do",
    "drop",
    "each",
    "else",
    "end",
    "escape",
    "except",
    "exclude",
    "exclusive",
    "exists",
    "explain",
    "fail",
    "filter",
    "first",
    "following",
    "for",
    "foreign",
    "from",
    "full",
    "generated",
    "glob",
    "group",
    "groups",
    "having",
    "if",
    "ignore",
    "immediate",
    "in",
    "index",
    "indexed",
    "initially",
    "inner",
    "insert",
    "instead",
    "intersect",
    "into",
    "is",
    "isnull",
    "join",
    "key",
    "last",
    "left",
    "like",
    "limit",
    "match",
    "materialized",
    "natural",
    "no",
    "not",
    "nothing",
    "notnull",
    "null",
    "nulls",
    "of",
    "offset",
    "on",
    "or",
    "order",
    "others",
    "outer",
    "over",
    "partition",
    "plan",
    "pragma",
    "preceding",
    "primary",
    "query",
    "raise",
    "range",
    "recursive",
    "references",
    "regexp",
    "reindex",
    "release",
    "rename",
    "replace",
    "restrict",
    "returning",
    "right",
    "rollback",
    "row",
    "rows",
    "savepoint",
    "select",
    "set",
    "table",
    "temp",
    "temporary",
    "then",
    "ties",
    "to",
    "transaction",
    "trigger",
    "unbounded",
    "union",
    "unique",
    "update",
    "using",
    "vacuum",
    "values",
    "view",
    "virtual",
    "when",
    "where",
    "window",
    "with",
    "without",
}



from abc import ABC, abstractmethod


class LLMIntegration(ABC):
    """This captures the integration of an LLM provider. It is designed to be
    subclassed and implemented for a specific LLM provider. It has one method, complete,
    which takes some untrusted input and returns the LLM's output, unfiltered."""

    @abstractmethod
    def complete(self, untrusted_input: str) -> str:
        """Send the untrusted input to the LLM and return the LLM's output.

        :param untrusted_input: The untrusted input from the user, but almost always
            wrapped in a :class:`PromptEnvelope <heimdallm.envelope.PromptEnvelope>`.
        :return: The untrusted output from the LLM.
        """
        raise NotImplementedError



from typing import Any, Optional, Sequence

from . import exc


class RequiredConstraint:
    """This represents a constraint that *must* be applied to the query.

    In the query, this comes in the form of ``table.column=:placeholder``. Enforced by
    the grammar, the comparison is always equality, the left hand side is always a
    fully-qualified column, and the right hand side is always a placeholder. These
    requirements ensure that the query is always constrained by a value that the
    developer specifies at query execution time.

    :param column: The fully-qualified column name.
    :param placeholder: The placeholder name for the value that your database expects to
        be interpolated at execution time.
    """

    def __init__(self, *, column: str, placeholder: str):
        self.fq_column = FqColumn.from_string(column)
        self.placeholder = placeholder

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, RequiredConstraint)
            and other.fq_column == self.fq_column
            and other.placeholder == self.placeholder
        )

    def __hash__(self):
        return hash((self.fq_column, self.placeholder))

    def __str__(self):
        return f"{self.fq_column}=:{self.placeholder}"

    __repr__ = __str__


class FqColumn:
    """This represents a fully-qualified column name.

    We require that LLM-produced queries from SQL Bifrosts use fully-qualified columns
    in their clauses, because if they didn't, we would need to infer which table owned
    the column, which requires runtime schema analysis. We could do that, but maybe in
    the future. It's much more straightforward to instruct that the LLM give us
    fully-qualified names.

    :param table: The table name.
    :param column: The column name.
    """

    __slots__ = ("table", "column")

    @classmethod
    def from_string(cls, fq_column_name: str) -> "FqColumn":
        """Parses a fully-qualified column name from a string, using the expected
        format of ``table.column``

        :param fq_column_name: The fully-qualified column name.
        :raises UnqualifiedColumn: If the string does not contain a period."""
        if "." not in fq_column_name:
            raise exc.UnqualifiedColumn(fq_column_name)
        table, column = fq_column_name.split(".")
        return cls(table=table, column=column)

    def __init__(self, *, table: str, column: str):
        self.table = table
        self.column = column

    def __str__(self):
        return f"{self.table}.{self.column}"

    @property
    def name(self) -> str:
        """A convenience property that returns the fully-qualified column name as a
        string in the same format ``table.column``"""
        return str(self)

    def __iter__(self):
        return iter((self.table, self.column))

    def __hash__(self):
        return hash((self.table, self.column))

    def __eq__(self, other):
        return self.table == other.table and self.column == other.column

    def __repr__(self):
        return f"{self.table}.{self.column}"


class JoinCondition:
    """This represents an equi-join between two tables, on two columns. The order of
    the fully-qualified columns does not matter; matching will work correctly in the
    code.

    :param first: The first fully-qualified column.
    :param second: The second fully-qualified column.
    :param identity: If the columns specified in this join condition can also be used as
        a :meth:`requester identity
        <heimdallm.bifrosts.sql.sqlite.select.validator.ConstraintValidator.requester_identities>`
        for the query, then this should be set to the name of the placeholder where the
        identity will be populated at runtime.
    """

    __slots__ = ("first", "second", "identity_placeholder")

    def __init__(self, first: str, second: str, *, identity: Optional[str] = None):
        self.first = FqColumn.from_string(first)
        self.second = FqColumn.from_string(second)
        self.identity_placeholder = identity

    @property
    def requester_identities(self) -> Sequence[RequiredConstraint]:
        """If this join condition has been marked as an identity join,
        construct the required constraints for both sides of the join. We'll use those
        constraints when testing for the requester's identity.

        :meta private:
        """
        if self.identity_placeholder:
            return [
                RequiredConstraint(
                    column=self.first.name,
                    placeholder=self.identity_placeholder,
                ),
                RequiredConstraint(
                    column=self.second.name,
                    placeholder=self.identity_placeholder,
                ),
            ]
        else:
            return []

    def __hash__(self):
        # it shouldn't matter which order the join condition comes in, the hash
        # should be the same
        return hash((self.first, self.second)) ^ hash((self.second, self.first))

    def __eq__(self, other):
        return (self.first == other.first and self.second == other.second) or (
            self.first == other.second and self.second == other.first
        )

    def __str__(self) -> str:
        return f"{self.first}={self.second}"

    __repr__ = __str__


class _AnyJoinCondition(JoinCondition):
    """A convenience class for representing any valid join condition."""

    def __init__(self):
        super().__init__("*.*", "*.*")


#: A convenience object that represents any valid join condition. Only use it for a
#: validator that represents full admin access to your database.
ANY_JOIN = _AnyJoinCondition()



from typing import TYPE_CHECKING, Optional

import lark

from heimdallm.support.github import make_ambiguous_parse_issue

if TYPE_CHECKING:
    from .common import FqColumn, JoinCondition, RequiredConstraint


class BaseException(Exception):
    """This is a convenience base class for all HeimdaLLM SQL exceptions to them easier
    to catch.

    :meta private:
    """


class GeneralParseError(BaseException):
    """
    Thrown from our grammar parser for anything that we want to bubble up as an
    InvalidQuery exception, but we don't have access to the raw input from within the
    parser
    """


class InvalidQuery(BaseException):
    """
    Thrown when our grammar cannot parse the unwrapped LLM output.

    :param query: The query that was attempted to be parsed.
    """

    def __init__(self, *, query: str):
        super().__init__(f"\n\n{query}\n")
        self.query = query


class ReservedKeyword(BaseException):
    """
    Thrown when a the query attempts to use a reserved keyword, unescaped, as an alias
    for a table or column.

    :param keyword: The reserved keyword that was used as an alias.
    """

    def __init__(self, *, keyword: str):
        super().__init__(f"Alias `{keyword}` is a reserved keyword")
        self.keyword = keyword


class AmbiguousParse(BaseException):
    """
    Thrown when our grammar parses the unwrapped LLM output, but the query results in
    multiple parse trees unexpectedly. This is a bug in our grammar, and it should be
    reported to the author via the link in the exception's output.

    :param trees: The list of parse trees that were generated from parsing the query.
    :param query: The query that was attempted to be parsed.
    :ivar issue_link: A link to the GitHub issue that should be opened to report this.
    """

    def __init__(self, *, trees: list[lark.ParseTree], query: str):
        self.issue_link = make_ambiguous_parse_issue(query, trees)

        super().__init__(
            f"Query resulted in {len(trees)} ambiguous parse trees. "
            "Please report this query to the HeimdaLLM maintainer using "
            f"the following link:\n{self.issue_link}"
        )
        self.trees = trees
        self.query = query


class UnqualifiedColumn(BaseException):
    """
    Thrown when a column isn't fully qualified in the form ``table.column``.

    :param column: The column that was not fully qualified.
    """

    def __init__(self, column: str):
        message = "Fully-qualified column name needs to be in the form 'table.column'"
        if column is not None:
            message += f" (got {column!r})"

        super().__init__(message)
        self.column = column


class IllegalSelectedColumn(BaseException):
    """
    Thrown when a column is selected that is not allowed by the constraint validator and
    it was not automatically removed because :doc:`/reconstruction` is disabled.

    :param column: The column that was selected. This is not a :class:`FqColumn
        <heimdallm.bifrosts.sql.common.FqColumn>` because we may not always have a table
        name.
    """

    def __init__(self, *, column: str):
        message = f"Column `{column}` is not allowed in SELECT"
        super().__init__(message)
        self.column = column


class IllegalConditionColumn(BaseException):
    """
    Thrown when a column is used in a ``JOIN`` condition or a ``WHERE`` condition that
    is not allowed by the constraint validator.

    :param column: The column that was used in the condition.
    """

    def __init__(self, *, column: "FqColumn"):
        message = f"Column `{column}` is not allowed in WHERE"
        super().__init__(message)
        self.column = column


class MissingRequiredConstraint(BaseException):
    def __init__(self, *, column: "FqColumn", placeholder: str):
        message = f"Missing required constraint `{column}`=:{placeholder}"
        super().__init__(message)
        self.column = column
        self.placeholder = placeholder


class MissingRequiredIdentity(BaseException):
    def __init__(self, *, identities: set["RequiredConstraint"]):
        message = f"Missing one required identities: {identities!r}"
        super().__init__(message)
        self.identities = identities


class IllegalJoinTable(BaseException):
    """
    Thrown when a join spec is not allowed by the constraint validator.

    :param join: The join spec that was not allowed.
    """

    def __init__(self, *, join: "JoinCondition"):
        message = f"Join condition {join} is not allowed"
        super().__init__(message)
        self.join = join


class IllegalJoinType(BaseException):
    """
    Thrown when a non-inner JOIN type is found.

    :param join_type: The type of JOIN that was found. This name comes directly from the
        grammar rule that captured it.
    """

    def __init__(self, *, join_type: str):
        message = f"JOIN type `{join_type}` is not allowed"
        super().__init__(message)
        self.join_type = join_type


class DisconnectedTable(BaseException):
    """
    Thrown when a table has joins, but the table in the ``FROM`` clause is not connected
    to any of those joins.

    :param table: The table that is not connected to a join.
    """

    def __init__(self, *, table: str):
        message = f"Table `{table}` is not connected to the query"
        super().__init__(message)
        self.table = table


class BogusJoinedTable(BaseException):
    """
    Thrown when a table's join condition does not include the table itself.

    :param table: The table that is not referenced in its own join condition.
    """

    def __init__(self, *, table: str):
        message = f"Join condition for `{table}` does not reference the table"
        super().__init__(message)
        self.table = table


class TooManyRows(BaseException):
    """
    Thrown when a query returns too many rows and :doc:`/reconstruction` is disabled.

    :param limit: The number of rows that the query wants to return. Nullable if no
        limit is specified.
    """

    def __init__(self, *, limit: Optional[int]):
        nice_limit = limit if limit is not None else "unlimited"
        message = f"Attempting to return too many rows ({nice_limit})"
        super().__init__(message)
        self.limit = limit


class IllegalFunction(BaseException):
    """
    Thrown when a disallowed SQL function has been used in the query.

    :param function: The lowercase name of the disallowed function.
    """

    def __init__(self, *, function):
        message = f"Function `{function}` is not allowed"
        super().__init__(message)
        self.function = function



import hashlib
from typing import Dict

from ..llm import LLMIntegration


class LookupMockLLM(LLMIntegration):  # pragma: no cover
    """a mock LLM integration that simply returns a canned response for the
    hash of some input"""

    def __init__(self) -> None:
        self.responses: Dict[str, str] = {}

    def complete(self, untrusted_user_input: str) -> str:
        hash_id = hashlib.md5(untrusted_user_input.encode("utf8")).hexdigest()
        return self.responses[hash_id]


class EchoMockLLM(LLMIntegration):  # pragma: no cover
    """a mock LLM integration that simply returns the provided response. useful
    in our tests where we know the query that we want to attempt to parse +
    validate."""

    def complete(self, sql_output: str) -> str:
        return sql_output



from typing import cast

from lark import Discard, Token
from lark import Transformer as _Transformer
from lark import Tree

from . import exc
from .common import FqColumn
from .utils.identifier import get_identifier, is_count_function
from .validator import ConstraintValidator
from .visitors.aliases import AliasCollector


def _build_limit_tree(limit, offset=None):
    children = [
        Token("LIMIT", "LIMIT"),
        Tree(
            "limit",
            [Token("NUMBER", limit)],
        ),
    ]
    if offset:
        children.extend(
            [
                Token("OFFSET", "OFFSET"),
                Tree(
                    "offset",
                    [Token("NUMBER", offset)],
                ),
            ]
        )
    return Tree("limit_clause", children)


def add_limit(limit_placeholder, max_limit: int):
    """ensures that a limit exists on the limit placeholder, and that it is not
    greater than the max limit"""

    # existing limit? test and maybe replace it
    if limit_placeholder.children:
        current_limit = int(
            cast(Token, next(limit_placeholder.find_data("limit")).children[0]).value
        )

        try:
            offset_node = next(limit_placeholder.find_data("offset"))
        except StopIteration:
            current_offset = 0
        else:
            current_offset = int(cast(Token, offset_node.children[0]).value)

        if current_limit > max_limit:
            limit_tree = _build_limit_tree(max_limit, current_offset)
            limit_placeholder.children[0] = limit_tree

    # adding a limit? just append it
    else:
        limit_tree = _build_limit_tree(max_limit)
        limit_placeholder.children.append(limit_tree)


class ReconstructTransformer(_Transformer):
    """makes some alterations to a query if it does not meet some basic validation
    constraints, but could with those alterations. currently, these are just the
    following:

        - adding or lowering a limit on the number of rows
        - removing illegal selected columns
    """

    def __init__(self, validator: ConstraintValidator, reserved_keywords: set[str]):
        self._validator = validator
        self._collector = AliasCollector(reserved_keywords=reserved_keywords)
        self._last_discarded_column: FqColumn | None = None
        self._reserved_keywords = reserved_keywords
        super().__init__()

    def transform(self, tree):
        self._collector.visit(tree)
        return super().transform(tree)

    def select_statement(self, children):
        """checks if a limit needs to be added or adjusted"""
        max_limit = self._validator.max_limit()

        if max_limit is not None:
            for child in children:
                if not isinstance(child, Tree):
                    continue

                if child.data == "limit_placeholder":
                    add_limit(child, max_limit)
                    break

        return Tree("select_statement", children)

    def selected_columns(self, children):
        # if there's no children, it means we discarded every column selected, meaning
        # that they were all illegal columns. since we can't proceed without a column,
        # go ahead and raise an exception about illegal column.
        if not children:
            raise exc.IllegalSelectedColumn(column=self._last_discarded_column.name)
        return Tree("selected_columns", children)

    def selected_column(self, children: list[Tree | Token]):
        """ensures that every selected column is allowed"""
        selected = children[0]
        if is_count_function(selected):
            pass

        elif isinstance(selected, Tree):
            for fq_column_node in selected.find_data("fq_column"):
                table_node, column_node = fq_column_node.children

                maybe_table_alias = get_identifier(table_node, self._reserved_keywords)
                column_name = get_identifier(column_node, self._reserved_keywords)

                table_name = self._collector._aliased_tables.get(
                    maybe_table_alias, maybe_table_alias
                )
                column = FqColumn(table=table_name, column=column_name)
                if not self._validator.select_column_allowed(column):
                    self._last_discarded_column = column
                    return Discard

        return Tree("selected_column", children)


def postproc(items):
    """helps format our reconstructed query so it's not all on one line"""
    for item in items:
        if isinstance(item, Token):
            if item.type in {
                "FROM",
                "GROUP_BY",
                "HAVING",
                "INNER_JOIN",
                "LIMIT",
                "OFFSET",
                "ORDER_BY",
                "WHERE",
                "WHERE_TYPE",
            }:
                yield "\n"
            else:
                pass

        yield item

        if isinstance(item, Token):
            if item.type in {}:
                yield " "



from abc import ABC, abstractmethod

from lark import Lark, ParseTree

from heimdallm.bifrost import Bifrost


class ConstraintValidator(ABC):
    """This is the base class for all constraint validators. It is used to validate
    the parse tree of the untrusted input, and to fix the parse tree if it fails a
    soft-validation pass."""

    @abstractmethod
    def fix(self, bifrost: Bifrost, grammar: Lark, tree: ParseTree) -> str:
        """If the tree fails validation in a soft-pass, attempt to reconstruct it
        to satisfy the constraints. If the tree cannot be reconstructed, throw a
        Bifrost-specific exception

        :param grammar: The Lark grammar used to parse the untrusted input. This is
            often used by :class:`lark.reconstruct.Reconstructor` to fix the parse tree.
        :param tree: The resulting parse tree of the untrusted input.
        """
        raise NotImplementedError

    @abstractmethod
    def validate(self, biforst: Bifrost, untrusted_input: str, tree: ParseTree):
        """This performs the Bifrost-specific validation of the parse tree. It throws a
        Bifrost-specific exception if the tree does not pass the constraints.

        :param untrusted_input: The untrusted input, which can be used to provide extra
            context if we need to raise an exception.
        :param tree: The resulting parse tree of the untrusted input.
        """
        raise NotImplementedError



from typing import cast

from lark import Token, Tree

from .. import exc


def get_identifier(node, reserved_keywords: set[str]) -> str:
    """takes some node from the tree, either a token or a subtree, and finds
    the identifier it contains. this is used for the rule that matches an
    identifier or a quoted identifier, because, in that case, an identifier will
    be a token, and a quoted identifier will be a subtree containing an
    identifier token.

    also we check to make sure that the identifier is not a reserved keyword, as
    a convenience. we could do it a separate step, but it's easier to do it
    here."""

    def match_ident(v):
        return isinstance(v, Token) and v.type == "IDENTIFIER"

    quoted = False
    ident = next(node.scan_values(match_ident)).value
    quoted = bool(list(node.find_data("quoted_identifier")))
    if not quoted and ident.lower() in reserved_keywords:
        raise exc.ReservedKeyword(keyword=ident)
    return ident


def is_count_function(node: Tree | Token) -> bool:
    """
    A helper for determining if a node is a *safe* aggregate function. We allow safe
    aggregate functions in the SELECT clause, because they don't reveal extra
    information.
    """
    if isinstance(node, Tree):
        if node.data == "aliased_column":
            node = node.children[0]

    if isinstance(node, Tree) and node.data == "function":
        fn_name = cast(Token, node.children[0].children[0]).value.lower()
        if fn_name == "count":
            return True
    return False



safe_date_functions = {
    "adddate",
    "addtime",
    "convert_tz",
    "curdate",
    "current_date",
    "current_time",
    "current_timestamp",
    "curtime",
    "date",
    "date_add",
    "date_format",
    "date_sub",
    "datediff",
    "day",
    "dayname",
    "dayofmonth",
    "dayofweek",
    "dayofyear",
    "extract",
    "from_days",
    "from_unixtime",
    "get_format",
    "hour",
    "last_day",
    "localtime",
    "localtimestamp",
    "makedate",
    "maketime",
    "microsecond",
    "minute",
    "month",
    "monthname",
    "now",
    "period_add",
    "period_diff",
    "quarter",
    "sec_to_time",
    "second",
    "str_to_date",
    "subdate",
    "subtime",
    "sysdate",
    "time",
    "time_format",
    "time_to_sec",
    "timediff",
    "timestamp",
    "timestampadd",
    "timestampdiff",
    "to_days",
    "to_seconds",
    "unix_timestamp",
    "utc_date",
    "utc_time",
    "utc_timestamp",
    "week",
    "weekday",
    "weekofyear",
    "year",
    "yearweek",
}

safe_string_functions = {
    "ascii",
    "bin",
    "bit_length",
    "char",
    "char_length",
    "character_length",
    "concat",
    "concat_ws" "elt",
    "export_set",
    "field",
    "find_in_set",
    "format",
    "from_base64",
    "hex",
    "insert",
    "instr",
    "lcase",
    "left",
    "length",
    # this is not safe
    # "load_file",
    "locate",
    "lower",
    "lpad",
    "ltrim",
    "make_set",
    "mid",
    "oct",
    "octet_length",
    "ord",
    "position",
    "quote",
    "regexp_instr",
    "regexp_like",
    "regexp_replace",
    "regexp_substr",
    "repeat",
    "replace",
    "reverse",
    "right",
    "rpad",
    "rtrim",
    "soundex",
    "space",
    "strcmp",
    "substr",
    "substring",
    "substring_index",
    "to_base64",
    "trim",
    "ucase",
    "unhex",
    "upper",
    "weight_string",
}

safe_math_functions = {
    "abs",
    "acos",
    "asin",
    "atan",
    "atan2",
    "ceil",
    "ceiling",
    "conv",
    "cos",
    "cot",
    "crc32",
    "degrees",
    "div",
    "exp",
    "floor",
    "ln",
    "log",
    "log10",
    "log2",
    "mod",
    "pi",
    "pow",
    "power",
    "radians",
    "rand",
    "round",
    "sign",
    "sin",
    "sqrt",
    "tan",
    "truncate",
}


safe_misc_functions = {
    "bit_count",
    "if",
    "ifnull",
    "nullif",
}

safe_json_functions = {
    "json_array",
    "json_array_append",
    "json_array_insert",
    "json_contains",
    "json_contains_path",
    "json_depth",
    "json_extract",
    "json_insert",
    "json_keys",
    "json_length",
    "json_merge",
    "json_merge_patch",
    "json_merge_preserve",
    "json_object",
    "json_overlaps",
    "json_pretty",
    "json_quote",
    "json_remove",
    "json_replace",
    "json_schema_valid",
    "json_schema_validation_report",
    "json_search",
    "json_set",
    "json_type",
    "json_unquote",
    "json_valid",
    "json_value",
}

safe_agg_functions = {
    "avg",
    "bit_and",
    "bit_or",
    "bit_xor",
    "count",
    "group_concat",
    "json_arrayagg",
    "json_objectagg",
    "max",
    "min",
    "std",
    "stddev",
    "stddev_pop",
    "stddev_samp",
    "sum",
    "var_pop",
    "var_samp",
    "variance",
}

safe_functions = (
    safe_date_functions
    | safe_string_functions
    | safe_math_functions
    | safe_misc_functions
    | safe_json_functions
    | safe_agg_functions
)

# https://dev.mysql.com/doc/mysqld-version-reference/en/keywords-8-0.html
reserved_keywords = {
    "accessible",
    "account",
    "action",
    "active",
    "add",
    "admin",
    "after",
    "against",
    "aggregate",
    "algorithm",
    "all",
    "alter",
    "always",
    "analyse",
    "analyze",
    "and",
    "any",
    "array",
    "as",
    "asc",
    "ascii",
    "asensitive",
    "at",
    "attribute",
    "authentication",
    "autoextend_size",
    "auto_increment",
    "avg",
    "avg_row_length",
    "backup",
    "before",
    "begin",
    "between",
    "bigint",
    "binary",
    "binlog",
    "bit",
    "blob",
    "block",
    "bool",
    "boolean",
    "both",
    "btree",
    "buckets",
    "bulk",
    "by",
    "byte",
    "cache",
    "call",
    "cascade",
    "cascaded",
    "case",
    "catalog_name",
    "chain",
    "challenge_response",
    "change",
    "changed",
    "channel",
    "char",
    "character",
    "charset",
    "check",
    "checksum",
    "cipher",
    "class_origin",
    "client",
    "clone",
    "close",
    "coalesce",
    "code",
    "collate",
    "collation",
    "column",
    "columns",
    "column_format",
    "column_name",
    "comment",
    "commit",
    "committed",
    "compact",
    "completion",
    "component",
    "compressed",
    "compression",
    "concurrent",
    "condition",
    "connection",
    "consistent",
    "constraint",
    "constraint_catalog",
    "constraint_name",
    "constraint_schema",
    "contains",
    "context",
    "continue",
    "convert",
    "cpu",
    "create",
    "cross",
    "cube",
    "cume_dist",
    "current",
    "current_date",
    "current_time",
    "current_timestamp",
    "current_user",
    "cursor",
    "cursor_name",
    "data",
    "database",
    "databases",
    "datafile",
    "date",
    "datetime",
    "day",
    "day_hour",
    "day_microsecond",
    "day_minute",
    "day_second",
    "deallocate",
    "dec",
    "decimal",
    "declare",
    "default",
    "default_auth",
    "definer",
    "definition",
    "delayed",
    "delay_key_write",
    "delete",
    "dense_rank",
    "desc",
    "describe",
    "description",
    "des_key_file",
    "deterministic",
    "diagnostics",
    "directory",
    "disable",
    "discard",
    "disk",
    "distinct",
    "distinctrow",
    "div",
    "do",
    "double",
    "drop",
    "dual",
    "dumpfile",
    "duplicate",
    "dynamic",
    "each",
    "else",
    "elseif",
    "empty",
    "enable",
    "enclosed",
    "encryption",
    "end",
    "ends",
    "enforced",
    "engine",
    "engines",
    "engine_attribute",
    "enum",
    "error",
    "errors",
    "escape",
    "escaped",
    "event",
    "events",
    "every",
    "except",
    "exchange",
    "exclude",
    "execute",
    "exists",
    "exit",
    "expansion",
    "expire",
    "explain",
    "export",
    "extended",
    "extent_size",
    "factor",
    "failed_login_attempts",
    "false",
    "fast",
    "faults",
    "fetch",
    "fields",
    "file",
    "file_block_size",
    "filter",
    "finish",
    "first",
    "first_value",
    "fixed",
    "float",
    "float4",
    "float8",
    "flush",
    "following",
    "follows",
    "for",
    "force",
    "foreign",
    "format",
    "found",
    "from",
    "full",
    "fulltext",
    "function",
    "general",
    "generate",
    "generated",
    "geomcollection",
    "geometry",
    "geometrycollection",
    "get",
    "get_format",
    "get_master_public_key",
    "get_source_public_key",
    "global",
    "grant",
    "grants",
    "group",
    "grouping",
    "groups",
    "group_replication",
    "gtid_only",
    "handler",
    "hash",
    "having",
    "help",
    "high_priority",
    "histogram",
    "history",
    "host",
    "hosts",
    "hour",
    "hour_microsecond",
    "hour_minute",
    "hour_second",
    "identified",
    "if",
    "ignore",
    "ignore_server_ids",
    "import",
    "in",
    "inactive",
    "index",
    "indexes",
    "infile",
    "initial",
    "initial_size",
    "initiate",
    "inner",
    "inout",
    "insensitive",
    "insert",
    "insert_method",
    "install",
    "instance",
    "int",
    "int1",
    "int2",
    "int3",
    "int4",
    "int8",
    "integer",
    "intersect",
    "interval",
    "into",
    "invisible",
    "invoker",
    "io",
    "io_after_gtids",
    "io_before_gtids",
    "io_thread",
    "ipc",
    "is",
    "isolation",
    "issuer",
    "iterate",
    "join",
    "json",
    "json_table",
    "json_value",
    "key",
    "keyring",
    "keys",
    "key_block_size",
    "kill",
    "lag",
    "language",
    "last",
    "last_value",
    "lateral",
    "lead",
    "leading",
    "leave",
    "leaves",
    "left",
    "less",
    "level",
    "like",
    "limit",
    "linear",
    "lines",
    "linestring",
    "list",
    "load",
    "local",
    "localtime",
    "localtimestamp",
    "lock",
    "locked",
    "locks",
    "logfile",
    "logs",
    "long",
    "longblob",
    "longtext",
    "loop",
    "low_priority",
    "master",
    "master_auto_position",
    "master_bind",
    "master_compression_algorithms",
    "master_connect_retry",
    "master_delay",
    "master_heartbeat_period",
    "master_host",
    "master_log_file",
    "master_log_pos",
    "master_password",
    "master_port",
    "master_public_key_path",
    "master_retry_count",
    "master_server_id",
    "master_ssl",
    "master_ssl_ca",
    "master_ssl_capath",
    "master_ssl_cert",
    "master_ssl_cipher",
    "master_ssl_crl",
    "master_ssl_crlpath",
    "master_ssl_key",
    "master_ssl_verify_server_cert",
    "master_tls_ciphersuites",
    "master_tls_version",
    "master_user",
    "master_zstd_compression_level",
    "match",
    "maxvalue",
    "max_connections_per_hour",
    "max_queries_per_hour",
    "max_rows",
    "max_size",
    "max_updates_per_hour",
    "max_user_connections",
    "medium",
    "mediumblob",
    "mediumint",
    "mediumtext",
    "member",
    "memory",
    "merge",
    "message_text",
    "microsecond",
    "middleint",
    "migrate",
    "minute",
    "minute_microsecond",
    "minute_second",
    "min_rows",
    "mod",
    "mode",
    "modifies",
    "modify",
    "month",
    "multilinestring",
    "multipoint",
    "multipolygon",
    "mutex",
    "mysql_errno",
    "name",
    "names",
    "national",
    "natural",
    "nchar",
    "ndb",
    "ndbcluster",
    "nested",
    "network_namespace",
    "never",
    "new",
    "next",
    "no",
    "nodegroup",
    "none",
    "not",
    "nowait",
    "no_wait",
    "no_write_to_binlog",
    "nth_value",
    "ntile",
    "null",
    "nulls",
    "number",
    "numeric",
    "nvarchar",
    "of",
    "off",
    "offset",
    "oj",
    "old",
    "on",
    "one",
    "only",
    "open",
    "optimize",
    "optimizer_costs",
    "option",
    "optional",
    "optionally",
    "options",
    "or",
    "order",
    "ordinality",
    "organization",
    "others",
    "out",
    "outer",
    "outfile",
    "over",
    "owner",
    "pack_keys",
    "page",
    "parser",
    "partial",
    "partition",
    "partitioning",
    "partitions",
    "password",
    "password_lock_time",
    "path",
    "percent_rank",
    "persist",
    "persist_only",
    "phase",
    "plugin",
    "plugins",
    "plugin_dir",
    "point",
    "polygon",
    "port",
    "precedes",
    "preceding",
    "precision",
    "prepare",
    "preserve",
    "prev",
    "primary",
    "privileges",
    "privilege_checks_user",
    "procedure",
    "process",
    "processlist",
    "profile",
    "profiles",
    "proxy",
    "purge",
    "quarter",
    "query",
    "quick",
    "random",
    "range",
    "rank",
    "read",
    "reads",
    "read_only",
    "read_write",
    "real",
    "rebuild",
    "recover",
    "recursive",
    "redofile",
    "redo_buffer_size",
    "redundant",
    "reference",
    "references",
    "regexp",
    "registration",
    "relay",
    "relaylog",
    "relay_log_file",
    "relay_log_pos",
    "relay_thread",
    "release",
    "reload",
    "remote",
    "remove",
    "rename",
    "reorganize",
    "repair",
    "repeat",
    "repeatable",
    "replace",
    "replica",
    "replicas",
    "replicate_do_db",
    "replicate_do_table",
    "replicate_ignore_db",
    "replicate_ignore_table",
    "replicate_rewrite_db",
    "replicate_wild_do_table",
    "replicate_wild_ignore_table",
    "replication",
    "require",
    "require_row_format",
    "reset",
    "resignal",
    "resource",
    "respect",
    "restart",
    "restore",
    "restrict",
    "resume",
    "retain",
    "return",
    "returned_sqlstate",
    "returning",
    "returns",
    "reuse",
    "reverse",
    "revoke",
    "right",
    "rlike",
    "role",
    "rollback",
    "rollup",
    "rotate",
    "routine",
    "row",
    "rows",
    "row_count",
    "row_format",
    "row_number",
    "rtree",
    "savepoint",
    "schedule",
    "schema",
    "schemas",
    "schema_name",
    "second",
    "secondary",
    "secondary_engine",
    "secondary_engine_attribute",
    "secondary_load",
    "secondary_unload",
    "second_microsecond",
    "security",
    "select",
    "sensitive",
    "separator",
    "serial",
    "serializable",
    "server",
    "session",
    "set",
    "share",
    "show",
    "shutdown",
    "signal",
    "signed",
    "simple",
    "skip",
    "slave",
    "slow",
    "smallint",
    "snapshot",
    "socket",
    "some",
    "soname",
    "sounds",
    "source",
    "source_auto_position",
    "source_bind",
    "source_compression_algorithms",
    "source_connect_retry",
    "source_delay",
    "source_heartbeat_period",
    "source_host",
    "source_log_file",
    "source_log_pos",
    "source_password",
    "source_port",
    "source_public_key_path",
    "source_retry_count",
    "source_ssl",
    "source_ssl_ca",
    "source_ssl_capath",
    "source_ssl_cert",
    "source_ssl_cipher",
    "source_ssl_crl",
    "source_ssl_crlpath",
    "source_ssl_key",
    "source_ssl_verify_server_cert",
    "source_tls_ciphersuites",
    "source_tls_version",
    "source_user",
    "source_zstd_compression_level",
    "spatial",
    "specific",
    "sql",
    "sqlexception",
    "sqlstate",
    "sqlwarning",
    "sql_after_gtids",
    "sql_after_mts_gaps",
    "sql_before_gtids",
    "sql_big_result",
    "sql_buffer_result",
    "sql_cache",
    "sql_calc_found_rows",
    "sql_no_cache",
    "sql_small_result",
    "sql_thread",
    "sql_tsi_day",
    "sql_tsi_hour",
    "sql_tsi_minute",
    "sql_tsi_month",
    "sql_tsi_quarter",
    "sql_tsi_second",
    "sql_tsi_week",
    "sql_tsi_year",
    "srid",
    "ssl",
    "stacked",
    "start",
    "starting",
    "starts",
    "stats_auto_recalc",
    "stats_persistent",
    "stats_sample_pages",
    "status",
    "stop",
    "storage",
    "stored",
    "straight_join",
    "stream",
    "string",
    "subclass_origin",
    "subject",
    "subpartition",
    "subpartitions",
    "super",
    "suspend",
    "swaps",
    "switches",
    "system",
    "table",
    "tables",
    "tablespace",
    "table_checksum",
    "table_name",
    "temporary",
    "temptable",
    "terminated",
    "text",
    "than",
    "then",
    "thread_priority",
    "ties",
    "time",
    "timestamp",
    "timestampadd",
    "timestampdiff",
    "tinyblob",
    "tinyint",
    "tinytext",
    "tls",
    "to",
    "trailing",
    "transaction",
    "trigger",
    "triggers",
    "true",
    "truncate",
    "type",
    "types",
    "unbounded",
    "uncommitted",
    "undefined",
    "undo",
    "undofile",
    "undo_buffer_size",
    "unicode",
    "uninstall",
    "union",
    "unique",
    "unknown",
    "unlock",
    "unregister",
    "unsigned",
    "until",
    "update",
    "upgrade",
    "url",
    "usage",
    "use",
    "user",
    "user_resources",
    "use_frm",
    "using",
    "utc_date",
    "utc_time",
    "utc_timestamp",
    "validation",
    "value",
    "values",
    "varbinary",
    "varchar",
    "varcharacter",
    "variables",
    "varying",
    "vcpu",
    "view",
    "virtual",
    "visible",
    "wait",
    "warnings",
    "week",
    "weight_string",
    "when",
    "where",
    "while",
    "window",
    "with",
    "without",
    "work",
    "wrapper",
    "write",
    "x509",
    "xa",
    "xid",
    "xml",
    "xor",
    "year",
    "year_month",
    "zerofill",
    "zone",
}



from lark import Transformer

from .. import exc
from ..utils.identifier import get_identifier


class AmbiguityResolver(Transformer):
    """this transformer's purpose is to resolve ambiguities in the parse tree
    that can only be resolved through some extra knowledge that would be
    difficult to embed in the grammar itself.
    """

    def __init__(self, query: str, reserved_keywords: set[str]):
        self.reserved_keywords = reserved_keywords
        self.query = query
        super().__init__()

    def test_alias(self, i, tree, trees) -> bool:
        """
        This resolves ambiguities in the parse tree related to aliases. For
        example, the following query is ambiguous:

            select t1.secret from t1 left join t2 on t1.jid = t2.jid

        Is "left" an alias for "t1", or is it part of the "left join"? It's
        ambiguous. We know it's not ambiguous because "left" is a keyword, but
        the parser can't know this. This class resolves those ambiguities by
        looking at all possible ambiguous parse trees in the ambiguity, and
        selecting only the one that does not have an alias conflict with a
        reserved keyword.
        """
        for alias_node in tree.find_data("generic_alias"):
            try:
                get_identifier(alias_node, self.reserved_keywords)
            except exc.ReservedKeyword:
                return False
        return True

    def test_req_comparisons(self, i, tree, trees) -> bool:
        """in a where_condition rule, the children can include
        `relational_comparison` and `required_comparison` rules, which are
        ambiguous, because required comparisons are a subset of relational
        comparisons. we always prefer to interpret the ambiguity as a required
        comparison though, because it is more strict, and it satisfies our
        required comparison validator constraints"""
        if tree.data in ("where_condition", "join_condition"):
            return tree.children[0].data == "required_comparison"
        return True

    def test_arith_expr(self, i, tree, trees) -> bool:
        """the arith_expr node is recursive, so it can be ambiguous. just choose the
        first parse of it, since this node doesn't really matter"""
        if tree.data == "arith_expr":
            return i == 0
        return True

    def _ambig(self, trees):
        def test_tree(i, tree):
            return (
                self.test_alias(i, tree, trees)
                and self.test_req_comparisons(i, tree, trees)
                and self.test_arith_expr(i, tree, trees)
            )

        pruned_trees = [tree for i, tree in enumerate(trees) if test_tree(i, tree)]

        if len(pruned_trees) == 0:
            raise exc.InvalidQuery(query=self.query)
        elif len(pruned_trees) == 1:
            return pruned_trees[0]
        else:
            raise exc.AmbiguousParse(trees=pruned_trees, query=self.query)



import jinja2

from heimdallm.bifrosts.sql.envelope import PromptEnvelope as _SQLPromptEnvelope


class PromptEnvelope(_SQLPromptEnvelope):
    __doc__ = _SQLPromptEnvelope.__doc__

    def template(self, env: jinja2.Environment) -> jinja2.Template:
        """
        Returns the template to use for the envelope. Override in a subclass for
        complete customization.

        :param env: The environment to use to load the template.
        :return: The template to use for the envelope.
        """
        return env.get_template("sql/mysql/select.j2")



from abc import abstractmethod
from itertools import chain
from typing import Optional, Sequence, cast

from lark import Lark, ParseTree
from lark.exceptions import VisitError
from lark.reconstruct import Reconstructor

from heimdallm.bifrost import Bifrost
from heimdallm.bifrosts.sql import exc
from heimdallm.bifrosts.sql.bifrost import Bifrost as _SQLBifrost
from heimdallm.constraints import ConstraintValidator as _BaseConstraintValidator

from .common import ANY_JOIN, FqColumn, JoinCondition, RequiredConstraint
from .visitors.aliases import AliasCollector
from .visitors.facets import FacetCollector, Facets


class ConstraintValidator(_BaseConstraintValidator):
    """
    This validator checks different of a SQL query. You are intended to derive this
    class and implement its methods."""

    @abstractmethod
    def requester_identities(self) -> Sequence[RequiredConstraint]:
        """Returns the possible identities of the requester, as represented in the
        database. This is used to instruct the LLM how to constrain the query that it
        generates. Only one of these identities needs to match for the query to be
        compliant.

        The reason that we return a sequence, and not a single identity, is that
        sometimes an LLM will specify the constraint as part of a ``JOIN`` condition,
        and not a ``WHERE`` condition. In that case, the column in the JOIN condition
        may not match the column you expect.

        For example, consider selecting films for a customer, constrained by the
        customer id. The LLM may give you a query like this:

        .. code-block:: sql

            SELECT f.title
            FROM film f
            JOIN inventory i ON f.film_id=i.film_id
            JOIN rental r ON i.inventory_id=r.inventory_id
            JOIN customer c ON r.customer_id=c.customer_id
            WHERE c.customer_id=:customer_id

        Or you may receive a query like this:

        .. code-block:: sql

            SELECT f.title
            FROM film f
            JOIN inventory i ON f.film_id=i.film_id
            JOIN rental r
                ON i.inventory_id=r.inventory_id
                AND r.customer_id=:customer_id

        Both ``rental.customer_id`` and ``customer.customer_id`` are valid requester
        identities, so ou need to specify both of them by returning a
        :class:`heimdallm.bifrosts.sql.common.RequiredConstraint` for each of them.

        :return: The sequence of possible requester identities.
        """
        raise NotImplementedError(
            "You must explicitly provide the requester identity, "
            "or an empty list for full access (dangerous)"
        )

    @abstractmethod
    def required_constraints(self) -> Sequence[RequiredConstraint]:
        """
        Returns a sequence of constraints that must exist in either the ``WHERE`` clause
        of the query or in a ``JOIN`` condition. It doesn't matter where the constraint
        is, as long as it exists and is required (i.e. not part of an optional
        condition).

        :return: The sequence of required constraints.
        """
        raise NotImplementedError(
            "You must explicitly provide a sequence of required constraints, "
            "or an empty list for no constraints"
        )

    @abstractmethod
    def select_column_allowed(self, column: FqColumn) -> bool:
        """Check that a fully-qualified column is allowed to be selected in the
        ``SELECT`` clause. Use this to restrict the columns and tables that can be
        selected.

        This value is also used to inform :doc:`/reconstruction`. Columns that do not
        pass this check will be removed from the query.

        :param column: The fully-qualified column.
        :return: Whether or not the column is allowed to be selected.
        """
        return False

    @abstractmethod
    def allowed_joins(self) -> Sequence[JoinCondition]:
        """Returns all of the tables allowed to be connected to the query via a
        ``JOIN`` and the equi-join conditions that must be met for the join to be valid.

        :return: The sequence of allowed joins.
        """
        return []

    @abstractmethod
    def max_limit(self) -> Optional[int]:
        """Return the maximum number of rows that can be returned by a query. if
        None, there is no limit.

        This value is also used to inform :doc:`/reconstruction`. If this function
        provides a limit, but the query does not, or the query provides a higher limit,
        the query will be reconstructed to include the correct limit.

        :return: The maximum number of rows that can be returned by a query, or None if
            unlimited.
        """
        return None

    @abstractmethod
    def can_use_function(self, function: str) -> bool:
        """
        Returns whether or not a SQL function is allowed to be used anywhere in the
        query. By default, this checks the function against the list of safe functions
        that we have curated by hand.

        :param function: The *lowercase* name of the function.
        :return: Whether or not the function is allowed.
        """
        raise NotImplementedError

    def condition_column_allowed(self, fq_column: FqColumn) -> bool:
        """
        Checks if a column is allowed to be used in a ``WHERE``, ``JOIN``, ``HAVING``,
        or ``ORDER BY``. By default, this calls :meth:`select_column_allowed`, but if
        you override this method and want to preserve that behavior, you should call
        yourself.

        :param fq_column: The fully-qualified column.
        :return: Whether or not the column is allowed to be used in a condition.
        """
        # let's default to "if you can see it, you can use it"
        return self.select_column_allowed(fq_column)

    def fix(self, bifrost: Bifrost, grammar: Lark, tree: ParseTree) -> str:
        """A parse tree may be valid SQL, but it may not be valid according to
        the validator's constraints. we may be able to make intelligent
        decisions about those constraints, and fix the parse tree though, for
        example, by adding a limit to a query.

        :meta private:
        """

        # gets around a circular import issue
        from heimdallm.bifrosts.sql import reconstruct

        transform = reconstruct.ReconstructTransformer(
            self,
            cast(_SQLBifrost, bifrost).reserved_keywords(),
        )
        try:
            fixed_tree = transform.transform(tree)
        except VisitError as e:
            if isinstance(e.orig_exc, exc.BaseException):
                raise e.orig_exc
            raise e

        output = Reconstructor(grammar).reconstruct(
            fixed_tree,
            postproc=reconstruct.postproc,
        )
        return output

    def validate(self, bifrost: Bifrost, untrusted_input: str, tree: ParseTree):
        """Analyze the parsed tree and validate it against our SQL constraints

        :param untrusted_input: The original query string. This is passed in so that if
            we need to raise an exception that references it, we can.
        :param tree: The parsed tree from the original query string.

        :meta private:
        """
        try:
            alias_collector = AliasCollector(
                cast(_SQLBifrost, bifrost).reserved_keywords()
            )
            alias_collector.visit(tree)

            facets = Facets()
            facet_collector = FacetCollector(
                facets,
                alias_collector,
                cast(_SQLBifrost, bifrost).reserved_keywords(),
            )
            facet_collector.visit(tree)
        except exc.GeneralParseError:
            raise exc.InvalidQuery(query=untrusted_input)

        # check the select column allowlist
        for fq_column in facets.selected_columns:
            if not self.select_column_allowed(fq_column):
                raise exc.IllegalSelectedColumn(column=fq_column.name)

        # check the join condition allowlist
        any_join_cond_allowed = ANY_JOIN in self.allowed_joins()
        for table, join_specs in facets.joined_tables.items():
            for join_spec in join_specs:
                allowed = any_join_cond_allowed or join_spec in self.allowed_joins()
                if not allowed:
                    raise exc.IllegalJoinTable(join=join_spec)

        # ensure that the selected table is joined to another table, if joins exist
        if (
            facets.joined_tables
            and not facets.joined_tables[cast(str, facets.selected_table)]
        ):
            raise exc.DisconnectedTable(table=cast(str, facets.selected_table))

        # ensure that all joins were joined on columns that exist on the joined tables
        if facets.bad_joins:
            raise exc.BogusJoinedTable(
                table=facets.bad_joins[0],
            )

        # all columns specified in the join conditions are automatically included in the
        # condition column allowlist
        allowed_join_conditions = set()
        for join_spec in self.allowed_joins():
            if join_spec == ANY_JOIN:
                continue
            allowed_join_conditions.add(join_spec.first)
            allowed_join_conditions.add(join_spec.second)

        # check the condition column allowlist
        for fq_column in facets.condition_columns:
            if fq_column in allowed_join_conditions:
                continue

            if not self.condition_column_allowed(fq_column):
                raise exc.IllegalConditionColumn(
                    column=fq_column,
                )

        # get all of the constraints that the query MUST be constrained by in
        # the WHERE clause
        for constraint in self.required_constraints():
            if constraint not in facets.required_constraints:
                raise exc.MissingRequiredConstraint(
                    column=constraint.fq_column,
                    placeholder=constraint.placeholder,
                )

        # verify that the query is constrained by at least one of the
        # requester's identities. this means that the results of the query will
        # be restricted to data that only the requester has access to.
        #
        # we look at both the identities returned explicitly from the
        # `requester_identities` method, and also the identities that have been
        # annotated in the join conditions.
        req_ids = self.requester_identities()
        join_identities = chain.from_iterable(
            j.requester_identities for j in self.allowed_joins()
        )
        all_req_identities = set(chain(req_ids, join_identities))
        if all_req_identities:
            found_id = False
            for one_id in all_req_identities:
                if one_id in facets.required_constraints:
                    found_id = True
                    break
            if not found_id:
                raise exc.MissingRequiredIdentity(identities=all_req_identities)

        # check that the query limits the rows correctly, if we restrict to a limit
        if (max_limit := self.max_limit()) is not None:
            if facets.limit is None or facets.limit > max_limit:
                raise exc.TooManyRows(limit=facets.limit)

        # check that every function used has been allowlisted
        for fn in facets.functions:
            if not self.can_use_function(fn):
                raise exc.IllegalFunction(function=fn)
